-- This file hooks addEventHandler (and removeEventHandler) and doesn't allow the specified events to trigger
-- until all clients have downloaded the client scripts.
-- Add a triggerServerEvent('onLoadedAtClient', getLocalPlayer()) to the onClientResourceStart
-- event for this to work.

g_EventHandlers = {
	onResourceStart = {},		-- { i = { elem = elem, fn = fn, getpropagated = bool } }
	onGamemodeMapStart = {},
	onPlayerJoin = {}
}

g_Root = getRootElement()
g_EventQueue = {}		-- { i = { name = eventName, source = source, args = {...} } }
g_NumLoadedPlayers = 0

addEventHandler('onResourceStart', getResourceRootElement(getThisResource()),
	function()
		g_QueueDisplay = textCreateDisplay()
		g_QueueTextItem = textCreateTextItem('Loading...', 0.5, 0.5, 'medium', 255, 255, 255, 255, 2, 'center', 'center')
		textDisplayAddText(g_QueueDisplay, g_QueueTextItem)
		for i,player in ipairs(getElementsByType('player')) do
			textDisplayAddObserver(g_QueueDisplay, player)
		end
	end
)

function updateQueueText()
	if g_QueueDisplay then
		local left = getPlayerCount() - g_NumLoadedPlayers
		textItemSetText(g_QueueTextItem, 'Waiting for ' .. left .. ' more player' .. (left ~= 1 and 's' or ''))
	end
end

function destroyQueueText()
	if g_QueueDisplay then
		textDestroyDisplay(g_QueueDisplay)
		g_QueueDisplay = nil
		textDestroyTextItem(g_QueueTextItem)
		g_QueueTextItem = nil
	end
end

addEvent('onLoadedAtClient', true)
addEventHandler('onLoadedAtClient', g_Root,
	function()
		g_NumLoadedPlayers = g_NumLoadedPlayers + 1
		if g_NumLoadedPlayers == getPlayerCount() then
			if g_QueueTimeoutTimer then
				killTimer(g_QueueTimeoutTimer)
			end
			flushQueue()
		else
			updateQueueText()
			if not g_QueueTimeoutTimer then
				g_QueueTimeoutTimer = setTimer(flushQueue, 15000, 1)
			end
		end
	end
)

addEventHandler('onPlayerQuit', g_Root,
	function()
		g_NumLoadedPlayers = g_NumLoadedPlayers - 1
		if g_NumLoadedPlayers == getPlayerCount() then
			if g_QueueTimeoutTimer then
				killTimer(g_QueueTimeoutTimer)
			end
			flushQueue()
		else
			updateQueueText()
		end
	end
)

function flushQueue()
	g_QueueTimeoutTimer = nil
	destroyQueueText()
	for _,event in ipairs(g_EventQueue) do
		source = event.source
		local triggeredElem
		for _,handler in ipairs(g_EventHandlers[event.name]) do
			triggeredElem = event.source
			if not triggeredElem then
				triggeredElem = g_Root
			end
			if isElement(triggeredElem) then
				while true do
					if triggeredElem == handler.elem then
						handler.fn(unpack(event.args))
						break
					end
					if not handler.getpropagated or triggeredElem == g_Root then
						break
					end
					triggeredElem = getElementParent(triggeredElem)
				end
			end
		end
	end
	g_EventQueue = {}
end

for eventName,_ in pairs(g_EventHandlers) do
	addEventHandler(eventName, g_Root, function(...) hookedEventTriggered(eventName, ...) end)
end

function hookedEventTriggered(eventName, ...)
	table.insert(g_EventQueue, { name = eventName, source = source, args = {...} })
	if g_NumLoadedPlayers == getPlayerCount() then
		if g_QueueTimeoutTimer then
			killTimer(g_QueueTimeoutTimer)
		end
		flushQueue()
	elseif eventName == 'onResourceStart' and not g_QueueTimeoutTimer then
		g_QueueTimeoutTimer = setTimer(flushQueue, 15000, 1)
	end
end

_addEventHandler = addEventHandler
function addEventHandler(event, elem, fn, getPropagated)
	if getPropagated == nil then
		getPropagated = true
	end
	if g_EventHandlers[event] then
		table.insert(g_EventHandlers[event], { elem = elem, fn = fn, getpropagated = getPropagated })
	else
		_addEventHandler(event, elem, fn, getPropagated)
	end
end

_removeEventHandler = removeEventHandler
function removeEventHandler(event, elem, fn)
	if g_EventHandlers[event] then
		local handler
		for i=#g_EventHandlers[event],1,-1 do
			handler = g_EventHandlers[event][i]
			if handler.elem == elem and handler.fn == fn then
				table.remove(g_EventHandlers[event], i)
			end
		end
	else
		_removeEventHandler(event, elem, fn)
	end
end